unit KnobsUtils;

{

  (C) COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Types, System.Classes, WinApi.Windows, System.Math, Vcl.Controls, Vcl.Forms, Vcl.Graphics;


const

  NullPoint : TPoint = ( x: 0; y: 0);

type

  TSample    = SmallInt;                            // External value type for OS samples
  TSamples   = packed array[ 0 .. MaxInt div SizeOf( TSample) - 1] of TSample;
  PSamples   = ^TSamples;

  TSValue    = Double;
  TSValues   = array of TSValue;
  PSValue    = ^TSValue;
  TSAngle    = Double; // For angles we we use the mathematical vision, 0 degrees
                       // points to the right, and going up means going counter
                       // clock wise.

  TDistanceMode = (
    dmCircular,
    dmHorizontal,
    dmVertical
  );


  function  ApplicationPath: String;
  procedure Exchange( Var A, B: Integer);
  function  RectanglesOverlap( A, B: TRect): Boolean;
  function  Clip( const aValue: TSValue; const aLow, aHigh: TSValue): TSValue;
  procedure RoundToStepValue( Var aValue: TSValue; aStepSize: TSValue);
  function  Distance( aPoint: TPoint): Integer;
  function  AddPoints( A, B: TPoint): TPoint;
  function  SubPoints( A, B: TPoint): TPoint;
  function  MultPoint( A: TPoint; M: Integer): TPoint;
  function  DivPoint( A: TPoint; D: Integer): TPoint;
  function  ValuesToAngle( MinValue, Value, MaxValue: TSValue): TSAngle;
  function  PolarToPoint( aCenter: TPoint; anAngle, aLength: TSAngle): TPoint;
  function  DistanceToValue( MinValue, MaxValue: TSValue; aPoint: TPoint; aDistanceMode: TDistanceMode): TSValue;
  function  Rotate( A: TPoint; anAngle: TSAngle): TPoint;
  function  ValueToStr( aValue: TSValue): string;
  function  IntegerToNoteStr( aNoteValue: Integer): string;
  function  BoolToValue( aValue : Boolean): TSValue;
  function  ValueToBool( aValue : TSValue): Boolean;

  function  Indent   ( anInd: Integer): string; overload;
  function  Indent   ( anInd: Integer; const S: string): string; overload;
  function  Indent   ( anInd: Integer; const aFmt: string; const anArgs: array of const): string; overload;

  procedure WriteInd  ( var aFile: TextFile; anInd: Integer; S: string);
  procedure WriteLnInd( var aFile: TextFile; anInd: Integer; S: string);

  function  MakePath( const aParts: array of string): string;
  function  Explode( const aString: string; const aChar: Char): TStringList;
  function  GetCpuClockCycleCount: Int64; assembler;

  procedure Nop;


implementation


var

  strIndent : string;


  // ---------------------------------------------------------------------------

  function ApplicationPath: string;
  begin
    Result := ExtractFilePath( Application.ExeName);
  end;


  // ---------------------------------------------------------------------------

  procedure Exchange( var A, B: Integer);
  var
    T : Integer;
  begin
    T := A;
    A := B;
    B := T;
  end;


  // ---------------------------------------------------------------------------

  function  RectanglesOverlap( A, B: TRect): Boolean;
  var
    R : TRect;
  begin
    Result := IntersectRect( R, A, B);
  end;


  // ---------------------------------------------------------------------------

  function Clip( const aValue: TSValue; const aLow, aHigh: TSValue): TSValue;
  begin
    Result := aValue;

    if aLow < aHigh
    then begin
      if aValue < aLow
      then Result := aLow
      else If aValue > aHigh
      then Result := aHigh;
    end
    else begin
      if aValue < aHigh
      then Result := aHigh
      else If aValue > aLow
      then Result := aLow;
    end;
  end;


  // ---------------------------------------------------------------------------

  procedure RoundToStepValue( Var aValue: TSValue; aStepSize: TSValue);
  begin
    if aStepSize <> 0
    then aValue := Round(( aValue + ( aStepSize / 2)) / aStepSize) * aStepSize;
  end;


  // ---------------------------------------------------------------------------

  function  Distance( aPoint: TPoint): Integer;
  begin
    with aPoint
    do Result := Round( Sqrt( x * x + y * y));
  end;


  // ---------------------------------------------------------------------------

  function   AddPoints( A, B: TPoint): TPoint;
  begin
    with Result
    do begin
      x := A.x + B.x;
      y := A.y + B.y;
    end;
  end;


  // ---------------------------------------------------------------------------

  function   SubPoints( A, B: TPoint): TPoint;
  begin
    with Result
    do begin
      x := A.x - B.x;
      y := A.y - B.y;
    end;
  end;


  // ---------------------------------------------------------------------------

  function   MultPoint( A: TPoint; M: Integer): TPoint;
  begin
    Result := Point( A.x * M, A.y * M);
  end;


  // ---------------------------------------------------------------------------

  function   DivPoint( A: TPoint; D: Integer): TPoint;
  begin
    Result := Point( A.x div D, A.y div D);
  end;


  // ---------------------------------------------------------------------------

  function  ValuesToAngle( MinValue, Value, MaxValue: TSValue): TSAngle;
  begin
    if MinValue <> MaxValue
    then Result := 270 * ( Value - MinValue) / ( MinValue - MaxValue) + 225
    else Result := 0; // arbitrary ...
  end;


  // ---------------------------------------------------------------------------

  function  PolarToPoint( aCenter: TPoint; anAngle, aLength: TSAngle): TPoint;
  begin
    with Result
    do begin
      x := Round( aCenter.x + aLength * Cos( DegToRad( anAngle)));
      y := Round( aCenter.y - aLength * Sin( DegToRad( anAngle)));
    end;
  end;


  // ---------------------------------------------------------------------------

  function  DistanceToValue( MinValue, MaxValue: TSValue; aPoint: TPoint; aDistanceMode: TDistanceMode): TSValue;
  var
    anAngle : Extended;
    anAlpha : Extended;
  begin
    case aDistanceMode of

      dmCircular :
        begin
          anAngle := RadToDeg( ArcTan2( - aPoint.y, aPoint.x));

          while anAngle < 0
          do anAngle := anAngle + 360;

          while anAngle > 270
          do anAngle := anAngle - 360;

          Result := Clip( MinValue + (( MinValue - MaxValue) * ( anAngle - 225) / 270), MinValue, MaxValue);
        end;

      dmHorizontal :
        begin
          anAlpha := 0.5 + aPoint.X / 200;
          Result  := Clip( MinValue + anAlpha * ( MaxValue - MinValue), MinValue, MaxValue);
        end;

      dmVertical :
        begin
          anAlpha := 0.5 - aPoint.Y / 200;
          Result  := Clip( MinValue + anAlpha * ( MaxValue - MinValue), MinValue, MaxValue);
        end;

      else begin
        Result := 0;
        Assert( False, 'unhandled knob control mode');
      end;
    end;
  end;


  // ---------------------------------------------------------------------------

  function   Rotate( A: TPoint; anAngle: TSAngle): TPoint;
  begin
    anAngle := DegToRad( anAngle);

    with Result
    do begin
      x := Round( Cos( anAngle) * A.x - Sin( anAngle) * A.y);
      y := Round( Sin( anAngle) * A.x + Cos( anAngle) * A.y);
    end;
  end;


  // ---------------------------------------------------------------------------

  function  ValueToStr( aValue: TSValue): string;
  begin
    Result := FloatToStrF( aValue, ffGeneral, 15, 0);
    // Result := Format( '%f', [ aValue]);
  end;


  // ---------------------------------------------------------------------------

  function  IntegerToNoteStr( aNoteValue: Integer): String;
  const
    Sharps : array[ 0 .. 11] Of Char = ' # #  # # # ';
    Notes  : array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
  var
    i : Integer;
  begin
    if ( aNoteValue < 0) or ( aNoteValue > 127)
    then Result := '??'
    else begin
      i := aNoteValue mod 12;
      Result := Format( '%s%d%s', [ Notes[ i], ( aNoteValue div 12) - 1, Sharps[ i]]);
    end;
  end;

  // ---------------------------------------------------------------------------

  function BoolToValue( aValue: Boolean): TSValue;
  begin
    if aValue
    then Result := 1.0
    else Result := 0.0;
  end;

  // ---------------------------------------------------------------------------

  function ValueToBool( aValue: TSValue): Boolean;
  begin
    Result := aValue > 0.5;
  end;


  // ---------------------------------------------------------------------------

  function  Indent( anInd: Integer): string; // overload;
  begin
    Result := Copy( strIndent, 1, 2 * anInd);
  end;


  // ---------------------------------------------------------------------------

  function  Indent( anInd: Integer; const S: string): string; // overload;
  begin
    Result := Format( '%s%s', [ Indent( anInd), S]);
  end;


  // ---------------------------------------------------------------------------

  function  Indent( anInd: Integer; const aFmt: string; const anArgs: array of const): string; // overload;
  begin
    Result := Indent( anInd, Format( aFmt, anArgs));
  end;


  // ---------------------------------------------------------------------------

  procedure WriteInd( var aFile: TextFile; anInd: Integer; S: string);
  begin
    Write( aFile, Indent( anInd), S);
  end;


  // ---------------------------------------------------------------------------

  procedure WriteLnInd( var aFile: TextFile; anInd: Integer; S: string);
  begin
    WriteLn( aFile, Indent( anInd), S);
  end;

  function  BuildStr( const aParts: array of string; const aSeparator: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := 0 to Length( aParts) - 1
    do begin
      if i = Length( aParts) - 1
      then Result := Result + aParts[ i]
      else Result := Result + aParts[ i] + aSeparator;
    end;
  end;

  function  MakePath( const aParts: array of string): string;
  begin
    Result := BuildStr( aParts, '/');
  end;

  function  Explode( const aString: string; const aChar: Char): TStringList;
  var
    p : Integer;
    i : integer;
    S : string;
  begin
    Result := TStringList.Create;
    p := 1;
    i := 1;
    while i <= Length( aString)
    do begin
      if aString[ i] = aChar
      then begin
        S := Copy( aString, p, i - p);
        Result.Add( S);
        p := i + 1;
      end;
      Inc( i);
    end;
    if p <= i
    then begin
      S := Copy( aString, p, Length( aString));
      Result.Add( S);
    end;
  end;


  function GetCpuClockCycleCount: Int64; assembler;
  // Reads 64-bit TSC counter running at CPU clock speed. From Torjei Kvinen
  asm
    dw $310F  // opcode for RDTSC
  end;

  procedure Nop;
  begin
    // a Nop ...
  end;


initialization

    strIndent := ' ';

    while Length( strIndent) < 256
    do strIndent := strIndent + strIndent;

finalization


end.
